import json
import uuid
from typing import Optional

import click
from tabulate import tabulate

from montecarlodata import settings
from montecarlodata.agents.fields import (
    EXPECTED_CREATE_OR_UPDATE_AGENT_RESPONSE_FIELD,
    EXPECTED_DELETE_AGENT_RESPONSE_FIELD,
    GCP_JSON_SERVICE_ACCOUNT_KEY,
    AWS_ASSUMABLE_ROLE,
)
from montecarlodata.common.common import read_as_json_string
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors, complain_and_abort
from montecarlodata.utils import GqlWrapper
from montecarlodata.queries.agent import (
    CREATE_OR_UPDATE_AGENT_MUTATION,
    DELETE_AGENT_MUTATION,
)


class AgentService:
    _AGENT_FRIENDLY_HEADERS = [
        "Agent ID",
        "Agent Type / Storage Type",
        "DC ID",
        "Endpoint",
        "Created on (UTC)",
        "Is Active",
    ]
    _VALIDATION_RESPONSE_HEADERS = [
        "Message",
        "Cause",
        "Resolution",
    ]

    def __init__(
        self,
        config: Config,
        request_wrapper: Optional[GqlWrapper] = None,
        user_service: Optional[UserService] = None,
    ):
        self._request_wrapper = request_wrapper or GqlWrapper(config)
        self._user_service = user_service or UserService(
            request_wrapper=self._request_wrapper, config=config
        )

    @manage_errors
    def create_agent(
        self, agent_type, platform, storage, auth_type, endpoint, **kwargs
    ) -> None:
        """
        Register an agent by validating connection and creating an AgentModel in the monolith.
        """

        agent_request = {
            "agent_type": agent_type,
            "endpoint": endpoint,
            "storage_type": storage,
            "platform": platform,
            "auth_type": auth_type,
        }

        if kwargs.get("dc_id"):
            agent_request["data_collector_id"] = kwargs["dc_id"]

        if auth_type == GCP_JSON_SERVICE_ACCOUNT_KEY:
            agent_request["credentials"] = read_as_json_string(kwargs["key_file"])
        elif auth_type == AWS_ASSUMABLE_ROLE:
            creds = {"aws_assumable_role": kwargs["assumable_role"]}
            if kwargs["external_id"]:
                creds["external_id"] = kwargs["external_id"]
            agent_request["credentials"] = json.dumps(creds)

        response = self._request_wrapper.make_request_v2(
            query=CREATE_OR_UPDATE_AGENT_MUTATION,
            operation=EXPECTED_CREATE_OR_UPDATE_AGENT_RESPONSE_FIELD,
            variables=agent_request,
        )

        agent_id = response.data.get("agentId")
        validation_result = response.data.get("validationResult")

        self._validate_response(validation_result)

        if agent_id is not None:
            click.echo("Agent successfully registered!\n" f"AgentId: {agent_id}")
        else:
            complain_and_abort("Failed to register agent.")

    @manage_errors
    def delete_agent(self, agent_id) -> None:
        """
        Deregister an Agent (deletes AgentModel from monolith)
        """
        agent_request = {"agentId": agent_id}

        response = self._request_wrapper.make_request_v2(
            query=DELETE_AGENT_MUTATION,
            operation=EXPECTED_DELETE_AGENT_RESPONSE_FIELD,
            variables=agent_request,
        )

        success = response.data.get("success")

        if success:
            click.echo(f"Agent {agent_id} deregistered.")
        else:
            complain_and_abort("Failed to deregister agent.")

    @manage_errors
    def echo_agents(
        self,
        headers: Optional[str] = "firstrow",
        table_format: Optional[str] = "fancy_grid",
    ):
        """
        Display agents in an easy to read table.
        """

        table = [self._AGENT_FRIENDLY_HEADERS]
        for agent in self._user_service.agents:
            is_active = "false" if agent.get("isDeleted") else "true"
            full_name = f"{agent.get('agentType', '')} / {agent.get('storageType', '')}"

            table += [
                [
                    agent.get("uuid", ""),
                    full_name,
                    agent.get("dc_id", ""),
                    agent.get("endpoint", ""),
                    agent.get("createdTime", ""),
                    is_active,
                ]
            ]

        # If the account has no agents, add 1 line of empty values so tabulate() creates a pretty empty table
        if len(table) == 1:
            table += ["" for _ in self._AGENT_FRIENDLY_HEADERS]

        click.echo(
            tabulate(table, headers=headers, tablefmt=table_format, maxcolwidths=100)
        )

    def _validate_response(self, validation_result):
        table = [self._VALIDATION_RESPONSE_HEADERS]
        stack_trace = None

        if validation_result.get("errors"):
            for error in validation_result.get("errors", []):
                stack_trace = (
                    "-----------" + "\n" + error.get("stackTrace", None) + "\n"
                )

                table += [
                    [
                        click.style(
                            f"ERROR: {error.get('friendlyMessage', '')}", fg="red"
                        ),
                        error.get("cause", ""),
                        error.get("resolution", ""),
                    ]
                ]

        if validation_result.get("warnings"):
            for warning in validation_result.get("warnings"):
                table += [
                    [
                        click.style(
                            f"WARNING: {warning.get('friendlyMessage', '')}",
                            fg="yellow",
                        ),
                        warning.get("cause", ""),
                        warning.get("resolution", ""),
                    ]
                ]

        # If there are any errors or warnings returned, display them
        if len(table) > 1:
            click.echo(
                tabulate(
                    table, headers="firstrow", tablefmt="fancy_grid", maxcolwidths=100
                )
            )

        # If Verbose Errors is set as env variable and there was an error stack trace, save trace to a file and print path
        if settings.MCD_VERBOSE_ERRORS and stack_trace:
            stack_trace_filename = f"mcd_error_trace_{uuid.uuid4()}.txt"
            with open(stack_trace_filename, "w") as stack_trace_file:
                stack_trace_file.write(stack_trace)
            click.echo(f"Stack Trace of error(s) saved to /{stack_trace_filename}")



import uuid
from typing import Optional, List, Dict, Any

import click
from pycarlo.common.errors import GqlError
from pycarlo.core import Client, Query
from pycarlo.lib.schema import TestCredentialsV2Response

from montecarlodata.common.echo_utils import (
    echo_warning_message,
    echo_error_message,
    styled_warning_icon,
    styled_error_icon,
    styled_success_icon,
)
from montecarlodata.common.user import UserService
from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort
from montecarlodata.utils import GqlWrapper


class CollectorValidationService:
    def __init__(
        self,
        config: Config,
        mc_client: Client,
        request_wrapper: Optional[GqlWrapper] = None,
        user_service: Optional[UserService] = None,
    ):
        self._mc_client = mc_client
        self._request_wrapper = request_wrapper or GqlWrapper(config)
        self._user_service = user_service or UserService(
            request_wrapper=self._request_wrapper, config=config
        )

    def run_validations(
        self, dc_id: Optional[str] = None, only_periodic: bool = False
    ) -> int:
        """
        Runs all validators for all integrations in the data collector.

        :param dc_id: The UUID of the data collector.
        :param only_periodic: Whether only periodic validations must be run or not.
        """
        dc = self._user_service.get_collector(dc_id)

        all_integrations = (
            (self._user_service.warehouses or [])
            + (self._user_service.bi_containers or [])
            + (self._user_service.etl_containers or [])
        )

        total_failures = 0
        for integration in all_integrations:
            click.echo("")
            connections: List[Dict[str, Any]] = integration.get("connections", [])
            for connection in connections:
                if "type" in connection and "uuid" in connection:
                    connection_failures = self._run_connection_validators(
                        dc_uuid=dc.uuid,
                        name=integration.get("name", ""),
                        connection_type=connection["type"].lower(),
                        connection_uuid=connection["uuid"],
                        only_periodic=only_periodic,
                    )
                    total_failures += connection_failures
        click.echo("")
        if total_failures == 0:
            click.echo("All validations passed")
        else:
            click.echo(f"{total_failures} validations failed")
        return total_failures

    def _run_connection_validators(
        self,
        dc_uuid: uuid.UUID,
        name: str,
        connection_type: str,
        connection_uuid: uuid.UUID,
        only_periodic: bool,
    ) -> int:
        click.echo(
            f"Validating {connection_type} connection: {name} - {connection_uuid}"
        )
        validations = self._get_supported_validations(
            dc_uuid=dc_uuid,
            connection_type=connection_type,
            only_periodic=only_periodic,
        )
        failures = 0
        for validation in validations:
            ok = False
            warning = False
            try:
                result = self._run_single_validation(
                    connection_uuid=connection_uuid,
                    validation_name=validation,
                )

                # if there are only warnings we log them but consider it a successful validation
                ok = result.success or not bool(result.errors)
                if result.warnings:
                    for warning in result.warnings:
                        echo_warning_message(f"\t{warning.cause}")
                if result.errors:
                    for error in result.errors:
                        echo_error_message(f"\t{error.cause}")
                elif not result.success:
                    warning = True
            except Exception as e:
                echo_error_message(f"\tValidation {validation} failed ({e}).")

            result_icon: str
            if ok:
                result_icon = (
                    styled_warning_icon() if warning else styled_success_icon()
                )
            else:
                result_icon = styled_error_icon()
                failures = failures + 1
            click.echo(f"\t{validation}: {result_icon}")

        return failures

    def _get_supported_validations(
        self, dc_uuid: uuid.UUID, connection_type: str, only_periodic: bool
    ) -> List[str]:
        query = Query()
        query.get_supported_validations_v2(
            dc_id=dc_uuid, connection_type=connection_type
        )

        try:
            result = self._mc_client(
                query=query,
            ).get_supported_validations_v2

            return [
                validation.name
                for validation in result.supported_validations
                if not only_periodic or validation.periodic_validation
            ]
        except GqlError as e:
            complain_and_abort(f"Unable to get list of supported validators ({e}).")
            return []

    def _run_single_validation(
        self,
        connection_uuid: uuid.UUID,
        validation_name: str,
    ) -> Optional[TestCredentialsV2Response]:
        query = Query()
        query.test_existing_connection_v2(
            connection_id=connection_uuid, validation_name=validation_name
        )

        return self._mc_client(
            query=query,
            idempotent_request_id=str(uuid.uuid4()),
            timeout_in_seconds=40,  # let Monolith timeout first
        ).test_existing_connection_v2



import click

from montecarlodata.agents.agent import AgentService
from montecarlodata.agents.fields import (
    DATA_STORE_AGENT,
    GCS,
    S3,
    GCP,
    AWS,
    AWS_ASSUMABLE_ROLE,
    GCP_JSON_SERVICE_ACCOUNT_KEY,
)
from montecarlodata.common.commands import DISAMBIGUATE_DC_OPTIONS
from montecarlodata.tools import add_common_options


@click.group(help="Manage a Monte Carlo Agent.")
def agents():
    """
    Group for any Agent related subcommands
    """
    pass


@agents.command(help="Register a Data Store Agent with remote S3 bucket.")
@click.pass_obj
@click.option(
    "--assumable-role",
    help="ARN of AWS assumable role.",
    required=True,
)
@click.option(
    "--bucket-name",
    help="Name of S3 bucket for data store.",
    required=True,
)
@click.option(
    "--external-id",
    help="AWS External ID.",
    required=False,
)
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
def register_s3_store(ctx, bucket_name, **kwargs):
    AgentService(config=ctx["config"]).create_agent(
        agent_type=DATA_STORE_AGENT,
        platform=AWS,
        storage=S3,
        auth_type=AWS_ASSUMABLE_ROLE,
        endpoint=bucket_name,
        **kwargs,
    )


@agents.command(help="Register a Data Store Agent with Google Cloud Storage.")
@click.pass_obj
@click.option(
    "--key-file",
    help="JSON Key file if auth type is GCP JSON service account key.",
    type=click.Path(exists=True),
    required=True,
)
@click.option(
    "--bucket-name",
    help="Name of GCS bucket for data store.",
    required=True,
)
@add_common_options(DISAMBIGUATE_DC_OPTIONS)
def register_gcs_store(ctx, bucket_name, **kwargs):
    AgentService(config=ctx["config"]).create_agent(
        agent_type=DATA_STORE_AGENT,
        platform=GCP,
        storage=GCS,
        auth_type=GCP_JSON_SERVICE_ACCOUNT_KEY,
        endpoint=bucket_name,
        **kwargs,
    )


@agents.command(help="Deregister an Agent.")
@click.pass_obj
@click.option("--agent-id", help="UUID of Agent to deregister.", required=True)
def deregister(ctx, agent_id):
    """
    Deregister an Agent (deletes AgentModel from monolith)
    """
    AgentService(config=ctx["config"]).delete_agent(agent_id)


@agents.command(help="List all agents in account.", name="list")
@click.pass_obj
def list_agents(ctx):
    """
    List all Agents in account
    """
    AgentService(config=ctx["config"]).echo_agents()



CREATE_OR_UPDATE_AGENT_MUTATION = """
mutation createOrUpdateAgent($authType:AuthTypeEnum!, $agentType:AgentTypeEnum!, $credentials:JSONString!, $endpoint:String!, $storageType:StorageTypeEnum!, $platform:PlatformTypeEnum, $upgradeable:Boolean, $wrapperType:String, $wrapperVersion:String, $imageVersion:String, $skipValidation:Boolean, $agentId:UUID, $dataCollectorId:UUID) {
  createOrUpdateAgent(authType:$authType, agentType:$agentType, credentials:$credentials, endpoint:$endpoint, platform:$platform, storageType:$storageType, dataCollectorId:$dataCollectorId, agentId:$agentId, skipValidation:$skipValidation, upgradeable:$upgradeable, wrapperType:$wrapperType, wrapperVersion:$wrapperVersion, imageVersion:$imageVersion) {
    agentId
    validationResult{
      success
      warnings{
        stackTrace
        friendlyMessage
        resolution
        cause
      }
      errors{
        stackTrace
        friendlyMessage
        resolution
        cause
      }
      validationName
      description
    }
  }  
}
"""

DELETE_AGENT_MUTATION = """
mutation deleteAgent($agentId:UUID!) {
  deleteAgent(agentId:$agentId) {
    success
  }
}
"""



# Query responses
EXPECTED_CREATE_OR_UPDATE_AGENT_RESPONSE_FIELD = "createOrUpdateAgent"
EXPECTED_DELETE_AGENT_RESPONSE_FIELD = "deleteAgent"

# Available platform types
AWS = "AWS"
GCP = "GCP"

# Available agent types
DATA_STORE_AGENT = "DATA_STORE_AGENT"
REMOTE_AGENT = "REMOTE_AGENT"

# Available authentication types
GCP_JSON_SERVICE_ACCOUNT_KEY = "GCP_JSON_SERVICE_ACCOUNT_KEY"
GCP_WORKLOAD_IDENTITY_FEDERATION = "GCP_WORKLOAD_IDENTITY_FEDERATION"
AWS_VPCE = "AWS_VPCE"
AWS_ASSUMABLE_ROLE = "AWS_ASSUMABLE_ROLE"

# Available storage types
GCS = "GCS"
S3 = "S3"

# Available wrapper types
TERRAFORM = "TERRAFORM"
